/* This file contains the functions that keep track of paths between hosts,
   and print them as needed. */

#include "netmap.h"
#include <string.h>

    int TrafficLW;
    static AbufSize;
#define AbufMAX 30000

/* ------------------------------------------------------------------------- */
/* Dump the path table that we have accumulated while reading the host file  */

dumppaths() {
    struct hpath *jpath;
    int		ih,jh,i,pw,pc,pdelta,
		nhib,			/* No. of hosts in bounds */
		layer;			/* color image layer, for priority */
    double	p1x,p1y,p2x,p2y,	/* Bezier control points */
    		Pwid,orgslope,destslope;
    char Abuf[AbufMAX];			/* buffer for white arrows */
    long CompassFlag;

	if (DebugLevel > 0) {
	    fprintf(stderr,"Examining %d hosts and %d paths in %d layers for inclusion...",
		NumHosts,RawPathCount,NumLayers);
	    if (Zmax > 0) fprintf(stderr,"Flow cutoff is %6.2lf\n",Cutoff);
	}
	for (layer=0; layer<=NumLayers; layer++) {
	    NumPaths=0; nhib = 0;
	    for (ih=1; ih<=NumHosts; ih++)
		if (htable[ih].inbounds != 0) nhib++;
    
	    fflush(stdout);
	    pw= 999; AbufSize = 0, pc=999;
	    for (i=1; i<=NumHosts; i++) {
		if (!htable[i].datacomplete) continue;
		jpath = htable[i].paths;
		while (jpath != NULL) {
    
 /* ih != i if this path has been removed */
		    ih = jpath->oindex;
		    jh = jpath->dindex;

 /*
   Check the color layer. Each time the color is changed, the layer
   number is incremented. Then we image the paths in order by layers.
   This causes the color overprinting to come out right.
  */ 
		    if (layer != jpath->passno) goto dontdraw;

    
/* Draw a path if either end is in the clipping region. Here we use the 
   "inbounds" flag as a 3-way instead of a Boolean. inbounds==0 means that
   the host is firmly out of bounds, and inbounds==1 means that it is
   in bounds. inbounds==2 means that it is on the edge, and that it got
   there by being moved from out of bounds. Only hosts that previously had
   inbounds==0 will ever get it set to 2.
 */
		    if ((ih > 0) && (jh > 0)) {
			if (((htable[ih].inbounds == 1)
			     || (htable[jh].inbounds == 1))
			     && htable[jh].datacomplete) {
			    if (htable[ih].inbounds == 0) movehost(ih,jh);
			    if (htable[jh].inbounds == 0) movehost(jh,ih);
    
	    
		    /* don't bother to recompute Pwid unless it has changed */
			    pdelta = 0;
			    if (pc != jpath->pcolor) {
				pc = jpath->pcolor;
				pdelta++;
			    }
			    if (pw != 0 && pw != jpath->pwidth) {
				pw = jpath->pwidth;
				if ((Zmax > 0)
				    && (!BiDirectional)
				    && (((100.0*pw)/Zmax) < Cutoff)) goto dontdraw;
				pdelta++;
			    }
			    if (pdelta) Pwid=setupLine(pw,pc);
    
    
/* Try to find a Bezier curve that goes between org and dest and that does
   not cross any other nodes. */
			    if (PlanPath(i,jh,htable[i].hstx,htable[i].hsty,
				     htable[jh].hstx,htable[jh].hsty,
				     &p1x,&p1y,&p2x,&p2y,(int) Pwid)) {
				PSP(htable[i].hstx,htable[i].hsty,"MT");
				PP(p1x,p1y); PP(p2x,p2y);
				PSP(htable[jh].hstx,htable[jh].hsty," CL");
				if (Zmax) PlaceArrows(htable[i].hstx,htable[i].hsty,
				     htable[jh].hstx,htable[jh].hsty,
				     p1x,p1y,p2x,p2y,Pwid,Abuf);
			    } else {
/* We can't find one. Just draw a straight line and hang the consequences.
   Since we called PlanPath, the values of p1 and p2 will be set properly
   (colinear in this case) for placing arrows and finding compass points. */
				PSP(htable[i].hstx,htable[i].hsty,"MT");
				PSP(htable[jh].hstx,htable[jh].hsty," PL");
				if (Zmax) PlaceArrows(htable[i].hstx,htable[i].hsty,
				     htable[jh].hstx,htable[jh].hsty,
				     p1x,p1y,p2x,p2y,Pwid,Abuf);
			    }
/* Having drawn the path, now mark at both ends the compass point at which
   the path came out or went in. This will make the job of PlanPath easier. */
			    orgslope = degs(atan2(p1y-htable[i].hsty,p1x-htable[i].hstx));
			    destslope = degs(atan2(p2y-htable[jh].hsty,p2x-htable[jh].hstx));
			    if (orgslope < 0) orgslope += 360.0;
			    if (destslope < 0) destslope += 360.0;
			    CompassFlag = 1 << (int) (orgslope/12.);
			    if (Pwid > 4) CompassFlag |= ((CompassFlag >> 1) | (CompassFlag << 1));
			    if (Pwid > 8) CompassFlag |= ((CompassFlag >> 1) | (CompassFlag << 1));
			    htable[i].cpoints |= CompassFlag;
			    CompassFlag = 1 << (int) (destslope/12.);
			    if (Pwid > 4) CompassFlag |= ((CompassFlag >> 1) | (CompassFlag << 1));
			    if (Pwid > 8) CompassFlag |= ((CompassFlag >> 1) | (CompassFlag << 1));
			    htable[jh].cpoints |= CompassFlag;
			    NumPaths++;
			} 
		    } 
    dontdraw:
		    if (jh > 0 && Zmax == 0) removepath(jh,i);
		    jpath = jpath->next;
		}
	    }
	    fputs(Abuf,stdout);
	}
    if (DebugLevel > 0)
        fprintf(stderr,"done.\nFound %d hosts and %d paths inside map boundaries.\n",
		nhib,NumPaths);
}


/* ----------------------------------------------------------------------- */

/* Move host "ih" so it is at the intersection of the clipping boundary and
   the line between it and host jh. */

movehost(ih,jh)
int ih,jh;
{
    double newx,newy;
    int vertclip;
    vertclip=atclip(htable[jh].hstx,htable[jh].hsty,
	   	    htable[ih].hstx,htable[ih].hsty,
	   	    &newx,&newy);
    if (vertclip < 0) return;
    if (SomeCircles) {
        htable[ih].drawn = DrawCircle;	/* mark to be drawn as circle */
        SlotLock(htable[ih].hname,&newx,&newy,vertclip ? 1 : -1);
 	if (vertclip)
	    newx = slotx(newx);
	else
	    newy = sloty(newy);
    }
    htable[ih].hstx = newx;
    htable[ih].hsty = newy;
    htable[ih].inbounds = 2;	/* flag it in the twilight zone */
    slot(newx,newy) = ih;
}

/* ----------------------------------------------------------------------- */

/* remove a redundant back path to prevent double-drawing of lines */
removepath (ih,jh) int ih,jh;
{
    struct hpath *jpath;
    jpath = htable[jh].paths;

    while (jpath != NULL) {
        if (((jpath->oindex)==jh) && ((jpath->dindex)==ih)) {
	    jpath->oindex=0;
	    jpath->dindex=0;
	}
	jpath = jpath->next;
    }
    jpath = htable[ih].paths;
    while (jpath != NULL) {
        if (((jpath->oindex)==ih) && ((jpath->dindex)==jh)) {
	    jpath->oindex=0;
	    jpath->dindex=0;
	}
	jpath = jpath->next;
    }
}

/* ----------------------------------------------------------------------- */


flushpaths ()
{   int ih,jh;
    struct hpath *ipath,*jpath;
    for (ih=1; ih<=NumHosts; ih++) {
        ipath = htable[ih].paths;
	if (htable[ih].datacomplete) {
	    jpath = ipath;
	    while (jpath != NULL) {
		jh= jpath->oindex;
		if (jh == ih) jh= jpath->dindex;
		if ( ! htable[jh].datacomplete ) {
		    jpath->oindex = 0;
		    jpath->dindex = 0;
		}
		jpath = jpath->next;
	    }
	} else {
	    while (ipath != NULL) {
		jpath = ipath->next;
		free(ipath);
		ipath = jpath;
	    }
	    htable[ih].paths = NULL;
	}
    }
}

/* ----------------------------------------------------------------------- */

/* This is a fragile hack function that examines a line path to see if it
   should be moved in some direction in order to avoid crossing other nodes.
   We try out various Bezier curves, and iterate along them to see whether
   they cross any nodes that they shouldn't cross. If they do, we reject
   that path and try a different Bezier.  If everything fails, we just
   draw a straight line between the 2 points, and to heck with it.

   */


int PlanPath(ih,jh,orgx,orgy,dstx,dsty,p1x,p1y,p2x,p2y,Pwid)
 double orgx,orgy,dstx,dsty,*p1x,*p1y,*p2x,*p2y;
{ 
    double 	ax,bx,cx,ay,by,cy,	/* Bezier control points */
    		side,			/* side bearing of trial curve */
		t,			/* x(t) and y(t) */
		xx,yy,			/* scratch */
		grainsize,		/* check at this interval */
		dist,			/* dist between the points */
	   	dx,dy,			/* its x and y components */
		deltat,
		orgslope,destslope,
		scf1x,scf1y,		/* intermediate values pulled out */
		scf2x,scf2y,		/* of Bezier loop for speed */
		zsin,zcos;		/* cos of path angle to x axis */
    int		i,sno,ix,jy,Aflag,iside,PTx;
    long	orgflag,destflag;

/* Here is the table of side bearings that we will try. */
    static float sides[] = {.05, .1, .15, .2, .3, .5, .75, 1.0, 1.25};
    static float side1[] = {1.0, -1.0,  0.5, -0.5, 1.0, 1.0, -1.0, -1.0, .25, .25, -.25, -.25};
    static float side2[] = {1.0, -1.0, -0.5,  0.5, .25,-.25,  .25, -.25, 1.0,-1.0, 1.0,  -1.0};
    static float PTA[] = {0.25, 0.0, -0.25};
    static float PTB[] = {0.75, 1.0,  1.25};
#define x(T) (orgx + T*(cx + T*(bx + T*ax)))
#define y(T) (orgy + T*(cy + T*(by + T*ay)))


#define PT1 0.25
#define PT2 0.75
    dx = dstx-orgx;   dy = dsty-orgy;
    dist = sqrt( dx*dx + dy*dy );
    scf1x = orgx+PT1*dx;    scf1y = orgy+PT1*dy;
    scf2x = orgx+PT2*dx;    scf2y = orgy+PT2*dy;
    *p1x = scf1x; *p1y = scf1y;
    *p2x = scf2x; *p2y = scf2y;
    
    if (dist < 3*Xslotinc) return(FALSE);
/*  if (!SomeCircles) return(FALSE);	*/
    zsin = dy/dist;    zcos = dx/dist;
    grainsize = dist / (Xslotinc/3);
    deltat = 1.0/grainsize;

    for (Aflag = 1; Aflag >= 0; Aflag--) {
        for (PTx = 0; PTx <= 2; PTx++) {
	    scf1x = orgx+PTA[PTx]*dx;    scf1y = orgy+PTA[PTx]*dy;
	    scf2x = orgx+PTB[PTx]*dx;    scf2y = orgy+PTB[PTx]*dy;
	    *p1x = scf1x; *p1y = scf1y;
	    *p2x = scf2x; *p2y = scf2y;
	    for (iside = 0; iside <= 11; iside++) {
		for(i=0; i<7; i++) {
		    side = sides[i]*dist;
		    *p1x = scf1x - side*side1[iside]*zsin;
		    *p1y = scf1y + side*side1[iside]*zcos;
		    *p2x = scf2x - side*side2[iside]*zsin;
		    *p2y = scf2y + side*side2[iside]*zcos;
		    cx = 3 * (*p1x - orgx);
		    cy = 3 * (*p1y - orgy);
		    bx = (((*p2x - orgx) - (*p1x - orgx)) * 3) - cx;
		    by = (((*p2y - orgy) - (*p1y - orgy)) * 3) - cy;
		    ax = dstx - orgx - cx - bx;
		    ay = dsty - orgy - cy - by;
		    orgslope = degs(atan2(*p1y-orgy,*p1x-orgx));
		    destslope = degs(atan2(dsty- *p2y,dstx- *p2x));
		    if (orgslope < 0) orgslope += 360.0;
		    if (destslope < 0) destslope += 360.0;
		    orgflag = Aflag << (int) (orgslope/12.);
		    destflag = Aflag << (int) (destslope/12.);
		    if (orgflag & htable[ih].cpoints) continue;
		    if (destflag & htable[jh].cpoints) continue;

		    for (t=0.05; t<0.95; t += deltat) {
			xx = x(t); yy = y(t);
			ix = xslot(xx); jy = yslot(yy);
			if (!clipxy(xx,yy)) goto nope1;
			sno = slotix(ix,jy);
			/* For wide curves, check surrounding slots as well. */
			if (Pwid > 7) sno |= slotix(ix+1,jy) | slotix(ix-1,jy) |
					     slotix(ix,jy+1) | slotix(ix,jy-1) |
				    slotix(ix+1,jy+1) | slotix(ix+1,jy-1) |
				    slotix(ix-1,jy+1) | slotix(ix-1,jy-1);
			if (sno && (sno == ih || sno == jh)) continue;
			if (sno) goto nope1;
		    }
		    if (side == 0.0) return(FALSE); return (TRUE);
       nope1: ;	
		}
	    }
	}
    }
    scf1x = orgx+PT1*dx;    scf1y = orgy+PT1*dy;
    scf2x = orgx+PT2*dx;    scf2y = orgy+PT2*dy;
    *p1x = scf1x; *p1y = scf1y;
    *p2x = scf2x; *p2y = scf2y;
    return(FALSE);
}

/* ----------------------------------------------------------------------- */

/*
 * This procedure places directional arrows on flow arcs, showing the 
 * direction of the traffic flow.
 *
 */

int PlaceArrows(orgx,orgy,dstx,dsty,p1x,p1y,p2x,p2y,Pwid,AA)
 double orgx,orgy,dstx,dsty,p1x,p1y,p2x,p2y,Pwid;
 char *AA;
{ 
    double 	ax,bx,cx,ay,by,cy,	/* Bezier control points */
		t,			/* x(t) and y(t) */
		xx,yy,			/* scratch */
		dist,			/* dist between the points */
	   	dx,dy,			/* its x and y components */
		dxdt,dydt,slope,
		lx,ldx;
    char	tbuf[100];

#define x(T) (orgx + T*(cx + T*(bx + T*ax)))
#define y(T) (orgy + T*(cy + T*(by + T*ay)))

    if (BiDirectional) return;
    dx = dstx-orgx;   dy = dsty-orgy;
    dist = sqrt( dx*dx + dy*dy ); /* distance in points */
    if (dist < 20.0) return;
    lx = 0.5; ldx = 100.;

/* If the vector is long enough and wide enough, give it 3 arrows  */
    if (dist > 72 && Pwid > 1.5) {lx = 0.25; ldx = 0.25;}

    for (t = lx; t < 1.0; t += ldx) {
	cx = 3 * (p1x - orgx);
	cy = 3 * (p1y - orgy);
	bx = (((p2x - orgx) - (p1x - orgx)) * 3) - cx;
	by = (((p2y - orgy) - (p1y - orgy)) * 3) - cy;
	ax = dstx - orgx - cx - bx;
	ay = dsty - orgy - cy - by;

	xx = x(t); yy = y(t);
	dxdt = 3*ax*t*t + 2*bx*t + cx;
	dydt = 3*ay*t*t + 2*by*t + cy;
	slope = degs(atan2(dydt,dxdt));
/* Use white arrows on fat lines and black arrows on skinny ones */
	if (Pwid <= 1.5) {
	    printf("%3.1f %3.1f ",Pwid > 0.8 ? 2.5 : 1.2,slope); PSP(xx,yy,"BA");
	    return;
	}
	sprintf(tbuf,"%3.1f %3.1f %3.1f %s\n",slope,xx,yy,
	     (Pwid > 4 ? "WA" : "SWA"));
	if (AbufSize + strlen(tbuf) > AbufMAX) {
	    fputs(AA, stdout);
	    strcpy(AA,(char *) NULL);
	    AbufSize = 0;
	}
	strcat(AA,tbuf);
	AbufSize += strlen(tbuf);
    }
}
