/* 
    This program reads a U.S. Census Bureau "picadad" tape and produces
    a Unix "dbm" file from it.

   Format of PICADAD file:

    1-5   5     Keypoint number
      6   1     check digit
    7-8   2     FIPS state code number (alphabetic)
			(see table below)
    9-11  3     FIPS county code number
   12-13  2     Census state code (geographic)
   			(see table below)
   14-15  2     USPS state code (alphabetic)
   			(see table below)
   16-35 20     Keypoint name
          1
      37  1	Place type
      			(C=city,T=town,V=village,B=borough,T=township,
			 Uunincorporated,Y=other)
	  1
      39  1	Place description digit
      			(1=incorporated central city of SMSA,
			 2=incorporated, population>2500
			 5=unincorporated, population>25000
			 6=Special Economic Urban Area
			 7=SEUA containing other places
			 8=New England town
			 9=none of the above)
	  1
      41  1	Part indicator (multicounty flag)
      			'P'==keypoint in more than one county
      42  1	Coextensive flag
      			'C' == city and county are identical
      		
   44-49  6	North latitude in decimal degrees.thousandths
   50-56  7	West longitude in decimal degrees.thousandths
   57-60  4	Census place code
   			(random 4-digit number)
      61  1	Size code
      			0	<2500		2	2.5k-5k
			3	5k-10k		4	10k-25k
			5	25k-50k		6	50k-100k
			7	100k-250k	8	250k-500k
			9	>500k
   62-65  4	SMSA number
   			(Stadard metropolitan statistical area)
      66  1	SMSA size
      			1	<250k		2	250k-500k
			3	500k-1m		3	>1m
   67-68  2	CTS manufacturing/production area
	  2
   71-73  3	BEA Economic area
	  3
   77-81  5	ZIP code or blank
	  1
      83  1	Keypoint digit
      			1=incorporated		2=unincorp > 25k
			3=New England town	4=tourist attraction
			5=border town		6=county center
			7=state center		8=misc
      84  1	Record type
      			(1=primary, 2=alternate, 3=spelling variant)


 */


#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <strings.h>
#include <sys/file.h>


#define	TRUE	1
#define	FALSE	0

int debuglevel=0;
char errbuf[256];



/* code name tables, indexed by FIPS state code number */

char stateNames[100][30];
char USPScodes[100][3];
int  censusCodes[100];

void st(censusCode,fipsCode,uspsCode,name)
int censusCode,fipsCode;
char *uspsCode,*name;
{
   censusCodes[fipsCode]=censusCode;
   strcpy(stateNames[fipsCode],name);
   strcpy(USPScodes[fipsCode],uspsCode);

}



main(argc, argv)
int	argc;
char	**argv;
{
	extern int optind;
	extern char *optarg;
	extern char *strcpy();
	int badopts=0;
	int c;

	while ((c = getopt(argc, argv, "d:")) != EOF)
	    switch (c) {

/* -d: debug. Argument is an integer giving debug level. */
		case 'd': sscanf(optarg,"%d",&debuglevel);
			  break;
		case '?': badopts++;
	    }
	fflush(stdout);
	if (badopts) {
	    exit (2);
	}
/* initialize state code tables */

	st(63,  1, "AL", "Alabama");
	st(94,  2, "AK", "Alaska");
	st(86,  4, "AZ", "Arizona");
	st(71,  5, "AR", "Arkansas");
	st(93,  6, "CA", "California");
	st(84,  8, "CO", "Colorado");
	st(16,  9, "CT", "Connecticut");
	st(51, 10, "DE", "Delaware");
	st(53, 11, "DC", "District of Columbia");
	st(59, 12, "FL", "Florida");
	st(58, 13, "GA", "Georgia");
	st(95, 15, "HI", "Hawaii");
	st(82, 16, "ID", "Idaho");
	st(33, 17, "IL", "Illinois");
	st(32, 18, "IN", "Indiana");
	st(42, 19, "IA", "Iowa");
	st(47, 20, "KS", "Kansas");
	st(61, 21, "KY", "Kentucky");
	st(72, 22, "LA", "Louisiana");
	st(11, 23, "ME", "Maine");
	st(52, 24, "MD", "Maryland");
	st(14, 25, "MA", "Massachusetts");
	st(34, 26, "MI", "Michigan");
	st(41, 27, "MN", "Minnesota");
	st(64, 28, "MS", "Mississippi");
	st(43, 29, "MO", "Missouri");
	st(81, 30, "MT", "Montana");
	st(46, 31, "NE", "Nebraska");
	st(88, 32, "NV", "Nevada");
	st(12, 33, "NH", "New Hampshire");
	st(22, 34, "NJ", "New Jersey");
	st(85, 35, "NM", "New Mexico");
	st(21, 36, "NY", "New York");
	st(56, 37, "NC", "North Carolina");
	st(44, 38, "ND", "North Dakota");
	st(31, 39, "OH", "Ohio");
	st(73, 40, "OK", "Oklahoma");
	st(92, 41, "OR", "Oregon");
	st(23, 42, "PA", "Pennsylvania");
	st(15, 44, "RI", "Rhode Island");
	st(57, 45, "SC", "South Carolina");
	st(45, 46, "SD", "South Dakota");
	st(62, 47, "TN", "Tennessee");
	st(74, 48, "TX", "Texas");
	st(87, 49, "UT", "Utah");
	st(13, 50, "VT", "Vermont");
	st(54, 51, "VA", "Virginia");
	st(91, 53, "WA", "Washington");
	st(55, 54, "WV", "West Virginia");
	st(35, 55, "WI", "Wisconsin");
	st(83, 56, "WY", "Wyoming");
	st(98, 98, "CN", "Canada");
	st(99, 99, "MX", "Mexico");


	while (optind<argc) {
	    fprintf(stderr,"Processing %s\n",argv[optind]);
	    processOneFile(argv[optind]);
	    optind++;
	}	
}

processOneFile(name)
char *name;
{
    FILE *inf;			/* input file */
    int nc;			/* number of characters read */
    char lbuf[128];		/* line buffer */
    char *fresult;		/* fgets result code */
    static char ss[256];	/* sscanf string */
    long    longSeconds;	/* longitude */
    long    latSeconds;		/* latitude  */
    int	    dalat,malat,salat;	/* deg,min,sec of lat */
    int	    dalon,malon,salon;	/* deg,min,sec of long */

	/* field variable definitions */
    int     keypointNumber;
    int     checkDigit;
    int     FIPSstateCode;
    int     FIPScountyCode;
    int     censusStateCode;
    char    USPSstateCode[3];
    char    keypointName[21];
    int     keypointType;
    int     placeDigit;
    int     partFlag;
    int     coExtensiveFlag;
    float   latitude;
    float   longitude;
    int     censusPlaceCode;
    int     sizeCode;
    int     SMSAnumber;
    int     SMSAsize;
    int     CTSarea;
    int     BEAarea;
    char    zipCode[6];
    int     kepointFlag;
    int     recordFlag;


    inf = fopen(name,"r");
    if (inf == NULL) {
        sprintf(errbuf,"Unable to open %s",name);
	perror(errbuf);
	return;
    }
    strcpy(ss,"");
    strcat(ss,"%5d");	/*  1-5   5     Keypoint number */
    strcat(ss,"%1c");	/*    6   1     check digit */
    strcat(ss,"%2d");	/*  7-8   2     FIPS state code number */
    strcat(ss,"%3d");	/*  9-11  3     FIPS county code number */
    strcat(ss,"%2d");	/* 12-13  2     Census state code (geographic) */
    strcat(ss,"%2c");	/* 14-15  2     USPS state code (alphabetic) */
    strcat(ss,"%20c");	/* 16-35 20     Keypoint name */
    strcat(ss,"%*1c");	/*        1 */
    strcat(ss,"%1c");	/*    37  1	Keypoint type  */
    strcat(ss,"%*1c");	/*	  1 */
    strcat(ss,"%1c");	/*    39  1	Place description digit */
    strcat(ss,"%*1c");	/*	  1 */
    strcat(ss,"%1c");	/*    41  1	Part indicator (multicounty flag) */
    strcat(ss,"%1c");	/*    42  1	Coextensive flag */
    strcat(ss,"%*1c");  /*	  1 */
    strcat(ss,"%6f");	/* 44-49  6	North latitude */
    strcat(ss,"%7f");	/* 50-56  7	West longitude */
    strcat(ss,"%4d");	/* 57-60  4	Census place code */
    strcat(ss,"%1c");	/*    61  1	Size code or blank */
    strcat(ss,"%4d");	/* 62-65  4	SMSA number */
    strcat(ss,"%1c");	/*    66  1	SMSA size */
    strcat(ss,"%2d");	/* 67-68  2	CTS manufacturing/production area */
    strcat(ss,"%*2c");	/*	  2 */
    strcat(ss,"%3d");	/* 71-73  3	BEA Economic area */
    strcat(ss,"%*3c");	/*	  3 */
    strcat(ss,"%5c");	/* 77-81  5	ZIP code or blank */
    strcat(ss,"%*1c");	/*	  1 */
    strcat(ss,"%1c");	/*    83  1	Keypoint type digit */
    strcat(ss,"%1c");	/*    84  1	Record type */


    while (!feof(inf)) {
	fresult = fgets(lbuf,sizeof(lbuf),inf);
	if (fresult == NULL) {
	    if ferror(inf) {
	        sprintf(errbuf,"Error reading %s",name);
		perror(errbuf);
	    }
	    break;
	}
	nc=strlen(lbuf);
	if (nc<84) {
	    if (nc>0)
	        fprintf(stderr,"Short record, nc=%d, ignored\n",nc);
		fprintf(stderr,"%s\n",lbuf);
		continue;
	}
	bzero(USPSstateCode,sizeof(USPSstateCode));
	bzero(keypointName,sizeof(keypointName));
	bzero(zipCode,sizeof(zipCode));
	sscanf(lbuf,ss,
		&keypointNumber,
		&checkDigit,
		&FIPSstateCode,
		&FIPScountyCode,
		&censusStateCode,
		USPSstateCode,
		keypointName,
		&keypointType,
		&placeDigit,
		&partFlag,
		&coExtensiveFlag,
		&latitude,
		&longitude,
		&censusPlaceCode,
		&sizeCode,
		&SMSAnumber,
		&SMSAsize,
		&CTSarea,
		&BEAarea,
		zipCode,
		&kepointFlag,
		&recordFlag);

	latSeconds  = (long) (3600.0 * latitude);
	longSeconds = (long) (3600.0 * longitude);
	
	dalat = latSeconds/3600;
	malat = (latSeconds / 60) % 60;
	salat = latSeconds % 60;

 /* longitude value is West, which should be negative, except that we know
    they are all in this hemisphere, so we can finesse it. */

	dalon = longSeconds/3600;
	malon = (longSeconds / 60) % 60;
	salon = longSeconds % 60;

	printf("%20s %2s %d %d %d N  %d %d %d W\n",
		keypointName,USPSstateCode,
		dalat,malat,salat,  dalon,malon,salon);
	
    }
}
